/**
* \file sudoku.c
*
* \brief Programme du jeu Sudoku execute en console.
*
* \author MORICEAU Tom
*
* \version 1.0
*
* \date 5 novembre 2023
*
* Jeu du sudoku éxécuté en console, avec une grille 
* de 9x9 cases. Le joueur doit remplir la grille en
* respectant les règles du sudoku.
*
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>


/*****************************************************
 * CONSTANTES ET TYPES *
*****************************************************/

/**
*
* \def N
*
* \brief constante qui definit le nombre de blocs par ligne et par colonne 
*
*/
#define N 3

/**
* 
* \def TAILLE
*
* \brief constante qui definit la taille d'un bloc (N*N)
*
*/
#define TAILLE N*N

/**
*
* \typedef tGrille
*
* \brief type qui definit une grille de sudoku de taille TAILLE*TAILLE
*
* Le type tGrille est utilisé pour stocker les valeurs des cases de la grille.
*
*/
typedef int tGrille[TAILLE][TAILLE];


/*****************************************************
 * FONCTIONS ET PROCEDURES *
*****************************************************/

/**
*
* \fn void chargerGrille(tGrille g)
*
* \brief Procédure qui charge une grille de sudoku à partir d'un fichier binaire.
*
* \param g la grille à charger.
*
* \return rien.
*
* Cette procédure consiste à demander le nom du fichier à charger, puis à l'ouvrir si il existe.
*
*/
void chargerGrille(tGrille g){
    char nomFichier[30];
    FILE * f; /** création d'un pointeur de type FILE */ 

    printf("Nom du fichier ? "); 
    scanf("%s", nomFichier);
    f = fopen(nomFichier, "rb"); /** ouverture du fichier en mode lecture binaire */

    while (f==NULL){ /** tant qu'on ne trouve pas le fichier, on redemande le nom du fichier */
        printf("\n ERREUR sur le fichier %s\n", nomFichier);
        printf("Nom du fichier ?\n");
        scanf("%s", nomFichier);
        f = fopen(nomFichier, "rb"); 
    }

    fread(g, sizeof(int), TAILLE*TAILLE, f); /** lecture du fichier et stockage dans le tableau de type tGrille g */
    fclose(f); /** fermeture du fichier */
}

/**
*
* \fn void afficherGrille(tGrille grille)
*
* \brief Procédure qui affiche une grille de sudoku à partir d'un tableau de type tGrille.
*
* \param grille la grille à afficher.
*
* Consiste à afficher la grille de sudoku en parcourant le tableau grille par lignes puis par colonnes
* et en affichant les valeurs de chaque cases dans une chaîne de caractères. 
* On affichera un point si la case est vide.
*/
void afficherGrille(tGrille grille){
    int lignes, colonnes = 0; 
    printf("     1  2  3   4  5  6   7  8  9  \n   +---------+---------+---------+\n");  
    for (lignes = 0; lignes < TAILLE ; lignes++){ /** Parcours des lignes de la grille */
        printf("%d  | ", lignes+1); /** affichage du numero de ligne a gauche de la grille */
        for (colonnes = 0 ; colonnes < TAILLE ; colonnes++){ /** Parcours des colonnes de la grille */
            if((colonnes+1)%3 == 0){ /** si la colonne est 3,6 ou 9 (divisible par 3) */
                if(colonnes+1 == 3 || colonnes+1 == 6){ /** si la colonne est 3 ou 6 */
                    if(grille[lignes][colonnes] == 0){ 
                        printf(". | "); /** affichage d'un point si la case est vide */
                    }else { 
                        printf("%d | ", grille[lignes][colonnes]); /** affichage de la valeur de la case */
                    }
                }else { /** si la colonne est 9 */
                    if(grille[lignes][colonnes] == 0){
                        printf(". |\n");
                    }else{
                        printf("%d |\n", grille[lignes][colonnes]);
                    }
                }
            }else { /** si la colonne n'est pas 3,6 ou 9 */
                if(grille[lignes][colonnes] == 0){  
                    printf(".  "); 
                }else {
                    printf("%d  ", grille[lignes][colonnes]); 
                }
            }
       }
       if((lignes+1)%3 == 0){ /** toutes les 3 lignes, on envoie la ligne qui montre la separation des blocs */
            printf("   +---------+---------+---------+\n"); 
        }
   } 
}

/**
*
* \fn void saisir(int *entree)
*
* \brief Procédure qui vérifie la validité de la saisie du joueur puis stocke la valeur saisie dans un entier.
*
* \param entree la valeur saisie par le joueur (ligne, colonne ou valeur).
*
* Consiste à demander une saisie au joueur, puis à vérifier que la saisie 
* est bien un entier compris entre 1 et 9.
* Si la saisie n'est pas valide, on redemande une saisie au joueur.
* Si la saisie est valide, on stocke la valeur saisie dans un entier.
*/
void saisir(int *entree){
    /**
    *
    * \var char saisie[2]
    * 
    * \brief chaîne de caractères qui stocke la saisie du joueur avant de la convertir en entier.
    * 
    */
    char saisie[2];

    /**
    *
    * \var int valeur
    *
    * \brief entier qui stocke la valeur saisie par le joueur si la saisie est un entier compris entre 1 et 9.
    *
    */
    int valeur;
    
    scanf("%s", saisie);
    /** tant que la saisie n'est pas un entier compris entre 1 et 9, on redemande une saisie */
    while(sscanf(saisie, "%d", &valeur) == 0 || valeur > TAILLE || valeur <= 0 ){
        printf("\nLa valeur saisie est incorrecte, veuillez réessayer avec un entier compris entre 1 et 9 : ");
        scanf("%s", saisie);
    }
    *entree = valeur; /** on stocke la valeur saisie par le joueur dans entree */
}

/**
*
* \fn possible(tGrille grille, int nLigne, int nColonne, int val)
*
* \brief Fonction qui vérifie si la valeur saisie par le joueur est possible compte tenu des règles du sudoku.
*
* \param grille la grille de sudoku.
* \param nLigne la ligne choisie par le joueur.
* \param nColonne la colonne choisie par le joueur.
* \param val la valeur choisie par le joueur.
*
* \return un booleen qui indique si la valeur saisie est possible ou non.
*
* Consiste à vérifier si la valeur saisie par le joueur est déjà présente dans la ligne, la colonne ou le bloc.
*/
bool possible(tGrille grille, int nLigne, int nColonne, int val){
    /**
    *
    * \var bool estPossible
    *
    * \brief booleen qui indique si la valeur saisie est possible ou non d'après les règles du sudoku.
    * Si la valeur n'est ni dans la ligne, ni dans la colonne, ni dans le bloc, la valeur est possible.
    * Initialisée à true, elle devient false si la valeur est déjà présente dans la ligne, la colonne ou le bloc.
    * 
    */
    bool estPossible = true;

    /**
    * 
    * \var int ligneDuBloc
    * 
    * \brief entier qui stocke la ligne de départ du bloc dans lequel se trouve la case choisie par le joueur.
    * 
    */

    /**
    * \var int colonneDuBloc
    * 
    * \brief entier qui stocke la colonne de départ du bloc dans lequel se trouve la case choisie par le joueur.
    * 
    */
    int lignes, colonnes, ligneDuBloc, colonneDuBloc = 0;

    for (lignes = 0; lignes < TAILLE ; lignes++){  
        for (colonnes = 0 ; colonnes < TAILLE ; colonnes++){ /** double boucle, parcours valeur par valeur */
            if(lignes == nLigne && grille[lignes][colonnes] == val){ /** si la valeur est deja presente dans la ligne */
                estPossible = false; 
                printf("\nErreur : la valeur est déjà présente dans la ligne choisie\n");
            }
            if(colonnes == nColonne && grille[lignes][colonnes] == val){ /** si la valeur est deja presente dans la colonne */
                estPossible = false; 
                printf("\nErreur : la valeur est déjà présente dans la colonne choisie\n");
            }

    /**
    * \var int bloc[2]
    * 
    * \brief tableau qui stocke la ligne et la colonne du bloc dans lequel se trouve la case choisie par le joueur.
    * 
    */
    int bloc[2]; 

    /** on cherche la ligne de depart du bloc dans lequel se trouve la case choisie par le joueur 
     * en divisant la ligne choisie par le nombre de blocs par ligne
     */
    bloc[0] = lignes/N; 
    /** on cherche la colonne de depart du bloc dans lequel se trouve la case choisie par le joueur 
     * en divisant la colonne choisie par le nombre de blocs par colonne 
     */
    bloc[1] = colonnes/N;

    /** on parcourt le bloc dans lequel se trouve la case choisie par le joueur 
     * ligneDuBloc est calculée en multipliant la ligne de départ du bloc par le nombre de lignes par bloc
     * puis en soustrayant par le nombre de lignes par bloc - 1 (pour obtenir la ligne de départ du bloc)
     * 
     * colonneDuBloc est calculée en multipliant la colonne de départ du bloc par le nombre de colonnes par bloc
     * puis en soustrayant par le nombre de colonnes par bloc - 1 (pour obtenir la colonne de départ du bloc)
    */
    for(int ligneDuBloc = ((bloc[0]*N)-(N-1)); ligneDuBloc <= bloc[0]*N; ligneDuBloc++){
        for(int colonneDuBloc = ((bloc[1]*N)-(N-1)); colonneDuBloc <= bloc[1]*N; colonneDuBloc++){
            if(ligneDuBloc != lignes && colonneDuBloc != colonnes){ /** si on ne se trouve pas sur la case choisie par le joueur */
                if(grille[ligneDuBloc][colonneDuBloc] == val){ /** si la valeur est deja presente dans le bloc */
                    estPossible = false; 
                    printf("\nErreur : la valeur est déjà présente dans le bloc choisi\n");
                    
                }
            }
        }
    }
            
        }
    }


    return(estPossible); 
}

/**
*
* \fn grillePleine(tGrille grille)
*
* \brief Fonction qui vérifie si la grille est pleine ou non (ajoutée par mes soins)
*
* \param grille la grille de sudoku.
*
* \return un booleen qui indique si la grille est pleine ou non.
*
* Consiste à vérifier si la grille est pleine en parcourant la grille 
* et en vérifiant si toutes les cases sont remplies.
* Si une case est vide, la grille n'est pas pleine.
*/
bool grillePleine(tGrille grille){
    int lignes, colonnes = 0;
    bool estPleine = true;
    for (lignes = 0; lignes < TAILLE ; lignes++){
        for (colonnes = 0 ; colonnes < TAILLE ; colonnes++){
            if(grille[lignes][colonnes] == 0){ /** si on trouve une valeur vide en parcourant la grille */
                estPleine = false;
            }
        }
    }
    return(estPleine);
    }

/**
*
* \fn main()
*
* \brief Fonction principale du programme
*
* \return rien.
*
* Consiste à charger la grille, à afficher la grille, à demander au joueur de saisir une case et une valeur,
* à vérifier si la valeur saisie est possible, puis à afficher la grille à nouveau et redemander une saisie,
* jusqu'à ce que la grille soit pleine.
*/
int main(){
    /**
    * \var tGrille grille1
    * 
    * \brief grille de sudoku de taille TAILLE*TAILLE
    * 
    */ 
    tGrille grille1;
    int lignes;
    int colonnes;
    int valeur;

    chargerGrille(grille1);
    
    /** verification que la grille n'est pas pleine grace a une fonction qui renvoie un booleen */
    while(grillePleine(grille1) == false){ 
        afficherGrille(grille1);
       
        printf("Indices de la case ?");
        /** on teste la validite des saisies avec la procedure saisir */
        saisir(&lignes); 
        saisir(&colonnes); 
        if(grille1[lignes-1][colonnes-1] != 0){ /** si la case n'est pas vide */
            printf("IMPOSSIBLE, la case n'est pas libre.\n");
        }
        else{
            printf("Valeur à insérer ?");
            saisir(&valeur);
            /** on teste la validite de la valeur d'apres les regles du sudoku avec la fonction possible */
            if(possible(grille1, lignes-1, colonnes-1, valeur)){ 
                grille1[lignes-1][colonnes-1] = valeur; /** si la valeur est possible, on l'insere dans la grille */
            }
        }
    }
    
    printf("Grille pleine, fin de partie"); 
}